
#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <WiFiUdp.h>            
#include <Ticker.h>
#include "si5351.h"
#include <Wire.h>
#include <ESP_EEPROM.h>
#include <LiquidCrystal_I2C.h>

#include <stdint.h> 
#include <stddef.h>
#include <ctype.h>
#include <string.h>

LiquidCrystal_I2C lcd(0x27,20,4);

Ticker blinker;
Si5351 si5351; 

WiFiUDP          udp;
WiFiClient       client;
IPAddress        timeServerIP; 
WiFiServer       wifiServer(1357);
ESP8266WebServer server(80);

const char *ap_ssid = "BeaconSetup";  //Enter your wifi SSID
const char *ap_password = "passw0rd";  //Enter your wifi Password

char ssid[32];
char password[32];
char my_ip_addr[20];

IPAddress    static_ip(192, 168, 1, 200);  
IPAddress    gateway(192, 168, 1, 1);     
IPAddress    subnet(255, 255, 255, 0);

char           ntpServerName[32] = "0.0.0.0";
const int      NTP_PACKET_SIZE = 48;             
byte           packetBuffer[ NTP_PACKET_SIZE];  
unsigned int   localPort = 2390;   

unsigned long  dds_freq = 475700;
unsigned long  dds_step = 50;

unsigned long  wspr_freq = 475700;
unsigned long  cw_freq = 476200;
unsigned long  qrss_freq = 476200;
unsigned long  opera_freq = 478500;   // receive frequency 477.000 USB  (-1500 hz)

int  ore = 0;
int  minuti = 0;
int  secondi = 0;

// wspr defines
#define TONE_SPACING            146       // ~1.46 Hz
#define WSPR_CTC                10672     // CTC value for WSPR
#define SYMBOL_COUNT            162       // bits number
#define CORRECTION              0   
 
char    call[7] = "0";                         
char    loc[5] = "0";                            
uint8_t wspr_dbm = 0;
int     wspr_min = 0;
uint8_t tx_buffer[SYMBOL_COUNT];

int     FSKCW_DIFF = 5;
int     qrss_min = 5;
int     qrss_mode = 3;

#define  M_NONE   99
#define  M_CW     0
#define  M_QRSS   1
#define  M_FSKCW  2
#define  M_DFCW   3
#define  M_WSPR   4
#define  M_OPERA  5
#define  M_FREQ   6
#define  M_TUNE   7

#define NTP_RETRY_CNT  10

int   freq_cor = 0;
int   tx_mode = M_WSPR;
int   tx_mode_http = M_NONE;
int   tx_start_min = 0;
int   tx_count = 0;
int   tx_sync_val = 0;
bool  ntp_request_active = false;
bool  time_ok = false;
bool  client_connected = false;
bool  tx_active = false;
bool  clock_update = false;
bool  clock_display = false;
bool  tx_start = false;


#define KEY_MODE   D7   
#define KEY_ON     201
#define KEY_OFF    202

#define KEY_IAMBIC 100
#define KEY_VERT   101

#define KEY_NORM   301
#define KEY_INV    302

int  cw_delay = 100;
int  cw_key_mode = KEY_IAMBIC;
int  cw_key_rev = 0;

#define OPERA_BINS  239
char opera_bins[] = "00000000000000000000000000000000000000000000000"
                    "00000000000000000000000000000000000000000000000"
                    "00000000000000000000000000000000000000000000000"
                    "00000000000000000000000000000000000000000000000"
                    "00000000000000000000000000000000000000000000000"
                    "0000";
int opera_mode = 8;    
int opera_min = 0;                
                    
#define CW_ALPHA_MAX  26
char  cw_alpha[CW_ALPHA_MAX+1] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
char  cw_alpha_code[CW_ALPHA_MAX+1][6] =
                       {"13000", "31110", "31310", "31100", "10000", // a-e
                        "11310", "33100", "11110", "11000", "13330", // f-j
                        "31300", "13110", "33000", "31000", "33300", // k-o
                        "13310", "33130", "13100", "11100", "30000", // p-t
                        "11300", "11130", "13300", "31130", "31330", // u-y
                        "33110" }; 

#define CW_NUM_MAX  10                                   
const char  cw_num[CW_NUM_MAX+1] = "0123456789";
const char  cw_num_code[CW_NUM_MAX+1][6] = 
                     {"33333", "13333", "11333", "11133", "11113",
                      "11111", "31111", "33111", "33311", "33331" };

boolean DBG = true;

//-------------- config parms -------------
#define EEPROM_READ    200
#define EEPROM_WRITE   201

struct
{
  char qrss_msg[20] = "0";
  char qrss_min[8] = "0";
  char qrss_mode[3] = "0";
  char wspr_call[8] = "0";
  char wspr_locator[8] = "0";
  char wspr_dbm[8] = "0";       
  char wspr_min[8] = "0";                  // start minute
  char ssid_name[32] = "0";
  char ssid_pwd[32] = "0";
  char cw_key_mode[2] = "0";
  char cw_speed[4] = "0";
  char opera_bins[241] = "0";
  char opera_min[8] = "0";
  char opera_mode[3] = "0";
  char ntp_addr[32] = "0";
  char cw_rev[2] = "0";
  char freq_cor[8] = "0";
  char tx_sync_val[6] = "0";
  char tx_step[5] = "0";
} config_str;

#include "html_pages.h"
#include "ntp_code.h"
#include "wspr_code.h"

// ----------- ap mode code -----------
//---------------------------------
void setup_wifi_ap(void)
{
  WiFi.mode(WIFI_AP);
  WiFi.softAP(ap_ssid, ap_password);
  delay(100);

  IPAddress IP = WiFi.softAPIP();

  server.on("/", server_on_connect);
  server.on("/response", server_on_response);
  server.on("/back", server_on_back);
  server.onNotFound(server_error);
  server.begin();
  
  delay(100);
  
  while(true)
  {
    server.handleClient();
  }
     
}

//-------------------------
void server_error(void)
{
  const char s[] = R"rawliteral(
    <!DOCTYPE html>
    <html>
    <body>
    <h1>Invalid request</h1>
    </body>
    </html>  
  )rawliteral";

  server.send(200, "text/html", s);  
}

//----------------------------------------------------
void get_data_from_buffer(String s_f, char c[], int l)
{
  String s = "";
  
  if ( server.hasArg(s_f) && server.arg(s_f) != NULL )
  {
    s = server.arg(s_f);
  } else {
    s = "";
  }

  s.toCharArray(c, l);
      
}

//--------------------------
void save_data(void)
{
   char s_w[10] = "0";
   int  i;
   
   get_data_from_buffer("qrss_msg",     config_str.qrss_msg, 20);
   get_data_from_buffer("qrss_min",     config_str.qrss_min, 8);
   get_data_from_buffer("qrss_mode",    config_str.qrss_mode, 3);
   get_data_from_buffer("wspr_call",    config_str.wspr_call, 8);
   get_data_from_buffer("wspr_locator", config_str.wspr_locator, 8);
   get_data_from_buffer("wspr_dbm",     config_str.wspr_dbm, 8);
   get_data_from_buffer("wspr_min",     config_str.wspr_min, 8);
   get_data_from_buffer("ssid_name",    config_str.ssid_name, 32);
   get_data_from_buffer("ssid_pwd",     config_str.ssid_pwd, 32);
   get_data_from_buffer("cw_key_mode",  config_str.cw_key_mode, 2);
   get_data_from_buffer("cw_speed",     config_str.cw_speed, 4);
   get_data_from_buffer("opera_bins",   config_str.opera_bins, 240);
   get_data_from_buffer("opera_mode",   config_str.opera_mode, 2);
   get_data_from_buffer("opera_min",    config_str.opera_min, 8);
   get_data_from_buffer("ntp_addr",     config_str.ntp_addr, 32);
   get_data_from_buffer("cw_rev",       config_str.cw_rev, 2);
   get_data_from_buffer("freq_cor",     config_str.freq_cor, 8);
   get_data_from_buffer("tx_sync",      config_str.tx_sync_val, 6);
   get_data_from_buffer("tx_step",      config_str.tx_step, 5);
   
   get_data_from_buffer("tx_freq", s_w, 7);
   i = atoi(s_w);
   dds_freq = (unsigned long) i;
   display_dds_freq();
   
   eeprom_data(EEPROM_WRITE);
   delay(200);
   eeprom_data(EEPROM_READ);
   set_variables_with_eeprom_data();
  
}

//-------------------------------------------------
String html_replace(String s, String s_s, char c[])
{
  String s_w;
  String s_c;
   
  s_w = "\"";
  s_c = String(c);
  s_w = s_w + s_c;
  s_w = s_w + "\"";
  s.replace(s_s, s_w); 
  return(s);  
}

//----------------------------------------------------
String html_replace_1(String s)
{
  String s_w;

  s_w = s;
  
  switch(tx_mode)
  {
    case M_WSPR:   s_w.replace("r_m_1", "checked");
                   s_w.replace("r_m_2", ""); s_w.replace("r_m_3", ""); s_w.replace("r_m_4", "");
                   s_w.replace("r_m_5", ""); 
                   break;    
    case M_QRSS:   s_w.replace("r_m_2", "checked");
                   s_w.replace("r_m_1", ""); s_w.replace("r_m_3", ""); s_w.replace("r_m_4", "");
                   s_w.replace("r_m_5", "");  
                   break;               
    case M_DFCW:   s_w.replace("r_m_3", "checked");
                   s_w.replace("r_m_1", ""); s_w.replace("r_m_2", ""); s_w.replace("r_m_4", "");
                   s_w.replace("r_m_5", ""); 
                   break; 
                                     
    case M_FSKCW:  s_w.replace("r_m_4", "checked");
                   s_w.replace("r_m_1", ""); s_w.replace("r_m_2", ""); s_w.replace("r_m_3", "");
                   s_w.replace("r_m_5", ""); 
                   break;  

    case M_OPERA:  s_w.replace("r_m_5", "checked");
                   s_w.replace("r_m_1", ""); s_w.replace("r_m_2", ""); s_w.replace("r_m_3", "");
                   s_w.replace("r_m_4", ""); s_w.replace("r_m_4", ""); 
                   break;
                                                                                                                                
  }

  return(s_w);
    
}
  
//--------------------
void get_data(void)
{
  String s;
  char s_w[40];
   
  s = s_home;

  s = html_replace(s, "\"my_ssid\"", config_str.ssid_name);
  s = html_replace(s, "\"my_ssid_pwd\"", config_str.ssid_pwd);
  
  s = html_replace(s, "\"w_call\"", config_str.wspr_call);
  s = html_replace(s, "\"w_loc\"", config_str.wspr_locator);
  
  s = html_replace(s, "\"w_dbm\"", config_str.wspr_dbm);
  s = html_replace(s, "\"w_min\"", config_str.wspr_min);
 
  s = html_replace(s, "\"q_msg\"", config_str.qrss_msg);
  s = html_replace(s, "\"q_min\"", config_str.qrss_min);
  s = html_replace(s, "\"q_mod\"", config_str.qrss_mode);

  s = html_replace(s, "\"cw_k\"", config_str.cw_key_mode);
  s = html_replace(s, "\"cw_s\"", config_str.cw_speed);
  s = html_replace(s, "\"cw_r\"", config_str.cw_rev);

  s = html_replace(s, "\"o_b\"", config_str.opera_bins);
  s = html_replace(s, "\"o_m\"", config_str.opera_mode);
  s = html_replace(s, "\"o_min\"", config_str.opera_min);

  s = html_replace(s, "\"ntp_a\"", config_str.ntp_addr);
  s = html_replace(s, "\"fr_c\"", config_str.freq_cor);

  s = html_replace(s, "\"tx_s\"", config_str.tx_sync_val);
  s = html_replace(s, "\"tx_st\"", config_str.tx_step);
  
  dtostrf(dds_freq, 6, 0, s_w);
  s = html_replace(s, "\"tx_f\"", s_w);

  // ---- set time into the page ----
  sprintf(s_w, "%02d:%02d:%02d", ore, minuti, secondi);
  String s_t(s_w); 
  s.replace("b_time", s_w);

  s = html_replace_1(s); 
 
  server.send(200, "text/html", s); 
}

//---------------------------
void server_on_back(void)
{
  get_data();
}

//---------------------------
void server_on_response(void)
{
  String s_r = "";

  // ----- get tx mode -----
  if ( server.hasArg( "mode_type" ) && server.arg( "mode_type" ) != NULL )
  {
    s_r = server.arg( "mode_type" );
  }
  
  if (s_r.equals("mode_wspr"))  tx_mode_http = M_WSPR;  
  if (s_r.equals("mode_qrss"))  tx_mode_http = M_QRSS;  
  if (s_r.equals("mode_dfcw"))  tx_mode_http = M_DFCW;
  if (s_r.equals("mode_fskcw")) tx_mode_http = M_FSKCW;
  if (s_r.equals("mode_opera")) tx_mode_http = M_OPERA;   

  // --- get operation ----
  if ( server.hasArg( "op_type" ) && server.arg( "op_type" ) != NULL )
  {
    s_r = server.arg( "op_type" );
  }
  
  if (s_r.equals("save_data")) save_data(); 
  if (s_r.equals("get_data"))  get_data();   

  server.send(200, "text/html", config_updated);  
}

//------------------------
void server_on_connect(void)
{  
  get_data();
}

//--------------------------------------
void retrieve_send(int a, char s[])
{
  if (a < 10)
  {
    client.write('0');
    client.write(s[0]);
  } else {
    client.write(s[0]);
    client.write(s[1]);          
  }
}

//--------------------------
void retrieve_time(void)
{
  char  s[8];
  
  itoa(ore, s, 10);
  retrieve_send(ore, s);
  
  itoa(minuti, s, 10);
  retrieve_send(minuti, s);
  
  itoa(secondi, s, 10);
  retrieve_send(secondi, s);
}

//-------------------------------------------------------------
void lcd_two_strings(char s[], char s_1[], int blank, int wait)
{
  
  if (blank == 1) lcd.clear();
  lcd.home();
  lcd.setCursor(0, 0);
  lcd.print(s);
  lcd.setCursor(0, 1);
  lcd.print(s_1);

  delay(wait);
}

//--------------------------
void lcd_clear_line(int l)
{
  int  n;

  lcd.setCursor(0, l);
  for(n=0; n<16; n++)
  {
    lcd.print(' ');
  }
  
}

//---------------------------
void display_tx_on_off(int i)
{
  lcd.setCursor(12, 0);
      
  if (i == 1)
  { 
    lcd.print("TX");  
    tx_active = true;
  } else {
    lcd.print("  ");
    tx_active = false;
  }
}

//---------------------
void tx_opera(void)
{
  int   n, l;
  char  s[40];
  
  display_tx_on_off(1);
  
  for(n=0; n<OPERA_BINS; n++)
  {
     if (opera_bins[n] == '0')
     {
       si5351_set_freq(0); 
     } else { 
       si5351_set_freq(dds_freq); 
     }           

     switch(opera_mode)
     {
       case 1: delay(256); 
               break;
       case 2: delay(512); 
               break;    
       case 4: delay(1024); 
               break;   
       case 8: delay(2048); 
               break;          
     }

  }
  
  si5351_set_freq(0);  
  display_tx_on_off(0);
       
}


//---------------------
void tx_wspr(void)
{
  int      i;
  char     s[40];
  uint64_t l;
  
  display_tx_on_off(1);
  si5351.output_enable(SI5351_CLK0, 1);

  // Now do the rest of the message
  for(i=0; i<SYMBOL_COUNT; i++)
  {
      l = (dds_freq * 100) + (tx_buffer[i] * TONE_SPACING);
      si5351.set_freq(l, SI5351_CLK0);
      delay(683);           
  }
  
  si5351_set_freq(0);  
  display_tx_on_off(0);

}

//--------------------------------
void find_cw_code(char c, char s[])
{
  int  n;
  
  strcpy(s, "");
  
  if ((c >= 'A') & (c <= 'Z'))
  {
    for(n=0; n<CW_ALPHA_MAX; n++)
    {
      if (c == cw_alpha[n])
      {
        strncpy(s, cw_alpha_code[n], 5); 
        return;
      }    
    }  
  }
  
  if ((c >= '0') & (c <= '9'))
  {
    for(n=0; n<CW_NUM_MAX; n++)
    {
      if (c == cw_num[n])
      {
        strncpy(s, cw_num_code[n], 5);  
        return;
      }   
    }      
  }

  if (c == '/') strncpy(s, "31131", 5);
   
}

//-----------------------
void tx_qrss(void)
{
  unsigned long qrss_dit_len;
  int  n, l, m, rc;
  char c, d;
  char s[8];
 

  display_tx_on_off(1);
  si5351.output_enable(SI5351_CLK0, 1);

  qrss_dit_len = (qrss_mode * 1000);

  // low signal..
  if (tx_mode == M_FSKCW)
  { 
    si5351_set_freq(dds_freq);
    delay(qrss_dit_len*3);
  } else {
    si5351_set_freq(0);
  }

  // get number of characters 
  l = strlen(config_str.qrss_msg);
  
  for(n=0; n<l; n++)
  {
    c = config_str.qrss_msg[n];
    if (c == ' ')         // is a blank 
    {
      switch(tx_mode)
      {
        case M_QRSS: si5351_set_freq(0); 
                       delay(qrss_dit_len*6);
                       break;  
        case M_FSKCW : si5351_set_freq(dds_freq); 
                       delay(qrss_dit_len*6);
                       break;                                  
        case M_DFCW :  si5351_set_freq(0); 
                       delay(qrss_dit_len*3);
                       break;    
      }
      
      continue;
    }

    find_cw_code(c, s);

    for(m=0; m<5; m++)
    {
      if (s[m] == '0') break;

      // dot symbol
      if (s[m] == '1')
      {
        switch(tx_mode)
        {
          case M_FSKCW:  si5351_set_freq(dds_freq+FSKCW_DIFF);
                         delay(qrss_dit_len);
                         break;
          case M_DFCW:   si5351_set_freq(dds_freq+FSKCW_DIFF);
                         delay(qrss_dit_len);
                         break;   
          default:       si5351_set_freq(dds_freq);
                         delay(qrss_dit_len);                     
        }
      }

      // line symbol
      if (s[m] == '3')
      {
        switch(tx_mode)
        {
          case M_FSKCW:  si5351_set_freq(dds_freq+FSKCW_DIFF);
                         delay(qrss_dit_len*3);
                         break;
          case M_DFCW:   si5351_set_freq(dds_freq);
                         delay(qrss_dit_len);
                         break;   
          default:       si5351_set_freq(dds_freq);
                         delay(qrss_dit_len*3);                     
        }
      }

      // pause between symbols. Note for DFCW = 0.
      switch(tx_mode)
      {
        case M_QRSS:   si5351_set_freq(0); 
                       delay(qrss_dit_len);
                       break; 
        case M_FSKCW:  si5351_set_freq(dds_freq);
                       delay(qrss_dit_len); 
                       break;                              
      }
      
    }

    // pause between char
    if (tx_mode != M_DFCW)
    {
      delay(qrss_dit_len*2);
    } else {
      si5351_set_freq(0); 
      delay(qrss_dit_len);
    }

  }

  si5351_set_freq(0); 
  display_tx_on_off(0);

}

//---------------------------------------
void set_variables_with_eeprom_data(void)
{
  int  i;
  float  f, f_1;
    
  strcpy(ssid, config_str.ssid_name);
  strcpy(password, config_str.ssid_pwd);
  strcpy(call, config_str.wspr_call);
  strcpy(ntpServerName, config_str.ntp_addr);
  strcpy(loc, config_str.wspr_locator);
  
  i = atoi(config_str.wspr_dbm);
  wspr_dbm = (uint8_t) i;

  i = atoi(config_str.wspr_min);
  wspr_min =  i;

  i = atoi(config_str.qrss_min);
  qrss_min =  i;

  i = atoi(config_str.qrss_mode);
  qrss_mode =  i;

  if (config_str.cw_key_mode[0] == '0')
  {
    cw_key_mode = KEY_VERT;
  } else {
    cw_key_mode = KEY_IAMBIC;  
  }

  if (config_str.cw_rev[0] == '0')
  {
    cw_key_rev = KEY_NORM;
  } else {
    cw_key_rev = KEY_INV;  
  }


  i = atoi(config_str.cw_speed);
  cw_delay = (uint8_t) i;
  cw_delay = 60000 / (43 * cw_delay);  // 43 elements in the PARIS word..
  
  i = atoi(config_str.opera_mode);
  opera_mode =  i;

  i = atoi(config_str.opera_min);
  opera_min = i;

  i = atoi(config_str.freq_cor);
  freq_cor = i;

  i = atoi(config_str.tx_sync_val);
  tx_sync_val = i;

  i = atoi(config_str.tx_step);
  dds_step = (unsigned long) i;
   
}

//---------------------------
void eeprom_data(int flag)
{
  switch(flag)
  {
    case EEPROM_WRITE: EEPROM.put(0, config_str);
                       delay(200);
                       EEPROM.commit();
                       break;
    case EEPROM_READ:  EEPROM.get(0, config_str);
                       break;                  
  }
}

//-------------------------------------
void si5351_set_freq(unsigned long f)
{
  unsigned long l;

  if (f == 0)
  {
    si5351.output_enable(SI5351_CLK0, 0);
    return; 
  }

  l = f;
  l = l * 100;
  si5351.set_freq(l, SI5351_CLK0);
  si5351.output_enable(SI5351_CLK0, 1);
   
}



//--------------------------------------------------------------------
void ticker_interrupt()
{
  char  s[20];
  int   i;
  
  ++secondi;

  if (secondi == 60)
  {
    secondi = 0;
    ++ minuti;
    if (minuti == 60)
    {
      minuti = 0;
      ++ore;
      if (ore == 24)
      {
        ore = 0;
      }
    }
  }  

  if (clock_display == false) return;
  
  if (tx_active == false)
  {
    clock_update = true;
    sprintf(s, "%02d:%02d:%02d", ore, minuti, secondi);
    lcd.setCursor(0, 0);
    lcd.print(s);
    clock_update = false;

    // cw mode, freq_set, tune 
    if (tx_start_min == 0) return;
    
    if (secondi == 0)
    {
      i = minuti % tx_start_min;
      if (i == 0) tx_start = true;
    }  

  }
  
}



//-------------------------------
int check_time(int start_min)
{
  int  n, i, p;
  
  if (secondi == 0)
  {
    i = minuti % start_min;
    if (i == 0) return(1);
  }  

  return(0);
  
}

//------------------------------------
void wait_for_key_release(int key_num)
{
  int  i, i_1;

  while(true)
  {
    i = digitalRead(key_num);  
    delay(50);
    i_1 = digitalRead(key_num);
    if ((i == 1) & (i_1 == 1)) break; 
  }
  
}

//-------------------------------
int check_if_key_pressed(int key)
{
  int i, i_1;
  
  i = digitalRead(key);
  delay(50);
  i_1 = digitalRead(key);

  // de-bounce
  if (i == 1) return(KEY_OFF);
  if (i_1 == 1) return(KEY_OFF);
  
  return(KEY_ON);
}

//-----------------
void tune_rtn(void)
{
  int  i, rc;
  int  on_off = 0;
  char s[8];
  
  while(true)
  {
    rc = check_if_key_pressed(D6);
    if (rc == KEY_ON)
    {
      if (on_off == 0)
      {
        on_off = 1;
        si5351_set_freq(dds_freq); 
        display_tx_on_off(1);
      } else {
        on_off = 0;
        si5351_set_freq(0);
        display_tx_on_off(0);
      }
      wait_for_key_release(D6);  
    }
    
    rc = check_if_key_pressed(KEY_MODE);
    if (rc == KEY_ON) break;

    if (on_off == 1)
    {
      i = analogRead(A0);
      sprintf(s, "%4d", i);
      lcd.setCursor(10, 1);
      lcd.print(s);
      delay(100);
    } 
    
  }
  
}

//-----------------------
void cw_iambic_rtn(void)
{
  int  i;
  
  tx_active = true;
  
  while(true)
  {
    i = digitalRead(D5); 
    if (i == 0)
    {
      si5351_set_freq(dds_freq);
      if (cw_key_rev == KEY_NORM)
      {
        delay(cw_delay);
      } else {
        delay(cw_delay*3);  
      }
      si5351_set_freq(0);
      delay(cw_delay);
    } 

    i = digitalRead(D6); 
    if (i == 0)
    {
      si5351_set_freq(dds_freq);
      if (cw_key_rev == KEY_NORM)
      {
        delay(cw_delay*3);
      } else {
        delay(cw_delay);
      }
      si5351_set_freq(0);
      delay(cw_delay);
    } 
    
    i = digitalRead(D7);  
    if (i == 0) break;
    yield();
  }

  tx_active = false;
      
}


//--------------------
void cw_vert_rtn(void)
{
  int  i;
  
  tx_active = true;
  
  while(true)
  {
    i = digitalRead(D5); 
    if (i == 0)
    {
      si5351_set_freq(dds_freq);
      while(true)
      {
        yield();
        i = digitalRead(D5); 
        if (i == 1) break;
      }
      si5351_set_freq(0);
    } 
    
    i = digitalRead(D7);  
    if (i == 0) break;
    yield();
  }

  tx_active = false;
      
}

//--------------------
void wifi_rtn(void)
{
  int  i;

  while(true)
  {
    i = check_if_key_pressed(KEY_MODE); 
    if (i == KEY_ON) 
    {
      lcd.setCursor(0, 0);
      lcd.print("                ");
      return;
    }
    i = check_if_key_pressed(D6); 
    if (i == KEY_ON) break;
  }
  
  tx_active = true;

  lcd.setCursor(0, 1);
  lcd.print("Starting");
  delay(1000);
    
  server.on("/", server_on_connect);
  server.on("/response", server_on_response);
  server.on("/back", server_on_back);
  server.onNotFound(server_error);
  server.begin();

  lcd.setCursor(0, 0);
  lcd.print(my_ip_addr);
  lcd.setCursor(0, 1);
  lcd.print("Ready   ");
  
  while(true)
  {
    server.handleClient();
    
    i = check_if_key_pressed(KEY_MODE); 
    if (i == KEY_ON) break;
  }

  server.close();
  server.stop();

  lcd.setCursor(0, 0);
  lcd.print("                ");
      
  tx_active = false;
      
}


//--------------------------------
void display_dds_freq(void)
{
  char s[9];
  int   n;
  
  lcd.setCursor(0, 1);
  dtostrf(dds_freq, 6, 0, s);

  for(n=0; n<6; n++)
  {
    lcd.print(s[n]);
    if (n == 2) lcd.print(".");
  }
  
}

//---------------------------------
void check_dds_incr_decr(void)
{
  int  i;
  unsigned long step_val;

  step_val = dds_step;
  
  // increase the dds frequency
  i = check_if_key_pressed(D6);
  if (i == KEY_ON) 
  { 
    dds_freq = dds_freq + step_val;
    if (dds_freq > 479000) dds_freq = 479000;
    display_dds_freq();
    delay(300);
  }

  // decrease the dds frequency
  i = check_if_key_pressed(D5);
  if (i == KEY_ON) 
  { 
    dds_freq = dds_freq - step_val;
    if (dds_freq < 472000) dds_freq = 472000;
    display_dds_freq();
    delay(300);
  }  
}

//------------------------
void update_dds_freq(void)
{
  int i;
  
  while(true)
  {
    check_dds_incr_decr();

    // mode_key
    i = check_if_key_pressed(KEY_MODE);
    if (i == KEY_ON) break;
    
  }
  
}

//-----------------------
void display_mode(void)
{
  lcd.setCursor(10, 1);

  switch(tx_mode)
  {
    case M_CW:     lcd.print("CW    ");
                   break;
    case M_QRSS:   lcd.print("QRSS  ");
                   break;                          
    case M_WSPR:   lcd.print("WSPR  ");
                   break; 
    case M_FSKCW : lcd.print("CW_FSK");
                   break;       
    case M_DFCW  : lcd.print("CW_DFC");
                   break;                  
    case M_FREQ  : lcd.print("SET F ");
                   break;  
    case M_TUNE  : lcd.print("TUNE  ");
                   break;   
    case M_OPERA : lcd.print("OPERA ");
                   break;                                              
  }

}

//----------------------
void sync_ntp(void)
{
  int n, rc;
 
  
  for(n=0; n<NTP_RETRY_CNT; n++)
  {
    rc = send_ntp_request();
    if (rc == 1) break;
    delay(100);
  }
    
}

//------------------------------------------------------
void action(void (*call_func)(), unsigned long f, int t)  
{
  int  rc;

  if (f > 0) dds_freq = f;
  tx_start_min = t;
  
  display_dds_freq();
  display_mode();  
  
  while(true)
  {  
    server.handleClient();
    
    if (t > 0)
    {
      if (tx_start == true) 
      {
        call_func();
        ++tx_count;
        if (tx_count == tx_sync_val)
        {
          tx_count = 0;
          sync_ntp();
        }
      }
      tx_start = false;
    } else {
      call_func();
    }
    
    rc = check_if_key_pressed(KEY_MODE);
    if (rc == KEY_ON)
    {
       ++tx_mode;
       if (tx_mode > 7) tx_mode = 0;
       display_mode();
       wait_for_key_release(KEY_MODE);
       return;  
    }

    if (tx_mode == M_CW)   continue;
    if (tx_mode == M_FREQ) continue;
    if (tx_mode == M_TUNE) continue;

    // when in wspr and qrss mode check the frequency up/down keys
    check_dds_incr_decr();

    // web request to change mode
    if (tx_mode_http == M_NONE) continue;
    
    if (tx_mode_http != tx_mode)
    {
      tx_mode = tx_mode_http; 
      tx_mode_http = M_NONE;
      display_mode();
      return;
    }

    if (tx_mode_http == tx_mode)
    {
      tx_mode_http = M_NONE;
      return;
    }
    
  }

}

//-------------------------
void check_mode(void)
{
  int  i, rc;

  switch(tx_mode)
  {
    case M_WSPR:   action(tx_wspr, wspr_freq, wspr_min);
                   wspr_freq = dds_freq;
                   break;
                     
    case M_QRSS:   action(tx_qrss, qrss_freq, qrss_min);
                   qrss_freq = dds_freq;
                   break;
                     
    case M_FSKCW:  action(tx_qrss, qrss_freq, qrss_min);
                   qrss_freq = dds_freq;           
                   break;

    case M_DFCW:   action(tx_qrss, qrss_freq, qrss_min);
                   qrss_freq = dds_freq;           
                   break;               
                       
    case M_CW:     if (cw_key_mode == KEY_VERT)
                   {
                     action(cw_vert_rtn, cw_freq, 0);
                   } else {
                     action(cw_iambic_rtn, cw_freq, 0); 
                   }
                   break;       

    case M_OPERA:  action(tx_opera, opera_freq, opera_min);
                   opera_freq = dds_freq;
                   break;  

    case M_FREQ:   dds_freq = cw_freq;
                   action(update_dds_freq, 0, 0);
                   cw_freq = dds_freq;
                   break;               
                   
    case M_TUNE:   dds_freq = cw_freq;
                   action(tune_rtn, 0, 0);
                   cw_freq = dds_freq;
                   break;                                                                                 
  }
    
}

//----------------------------------------------------------------------
void wifi_connect(void)
{
  char s[30];
  int  i;
  
  lcd_two_strings("WiFi connect", ssid, 1, 1);    
  WiFi.begin(ssid, password);
  
  while (WiFi.status() != WL_CONNECTED) 
  {
    delay(1000);
  }

  //WiFi.config(static_ip, gateway, subnet);
  lcd_two_strings("WiFi connected", ssid, 1, 1000);    
   
  wifiServer.begin();
  udp.begin(localPort);

  sprintf(s, "%s", WiFi.localIP().toString().c_str() ); 
  lcd_two_strings("WiFi IP Address", s, 1, 4000);
  strcpy(my_ip_addr, s);
  
  lcd.clear();
}

//-------------
void setup() 
{
  int    rc, n, i;
  float  f;
  char   s[20];
   
  lcd.init();
  lcd.backlight();
  lcd.clear();
  lcd.noBlink(); 
  lcd_two_strings("TX/Beacon", "630M Band", 1, 2000);
  
  delay(200);

  pinMode(D5, INPUT_PULLUP);
  pinMode(D6, INPUT_PULLUP);
  pinMode(D7, INPUT_PULLUP);           // key mode

  EEPROM.begin(512);
  
  // if key pressed at startup we have to enter the setup dialog
  i = digitalRead(KEY_MODE);
  if (i == 0) 
  {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Setup");
    lcd.setCursor(0, 1);
    lcd.print("192.168.4.1");
    setup_wifi_ap();
  }
 
  eeprom_data(EEPROM_READ);
  set_variables_with_eeprom_data();     
  
  wspr_encode(call, loc, wspr_dbm, tx_buffer);

  // ---- connect to the local network -----
  wifi_connect(); 
  delay(1000);

  // --- start the web server -----
  server.on("/", server_on_connect);
  server.on("/response", server_on_response);
  server.on("/back", server_on_back);
  server.onNotFound(server_error);
  server.begin();
  
  // ---- start the timer and setup time ---- 
  blinker.attach(1, ticker_interrupt);

  lcd_two_strings("NTP Request", "", 1, 100);
  for(n=0; n<NTP_RETRY_CNT; n++)
  {
    lcd.setCursor(12, 0);
    lcd.print(n);
    rc = send_ntp_request();
    if (rc == 1) break;
    delay(100);
  }  

  lcd.setCursor(12, 0);
  if (time_ok == true)
  {
    lcd.print("OK");  
  } else{
    lcd.print("ERR"); 
    delay(5000);
  }

  delay(2000);
  lcd.clear();
  clock_display = true;
 
  // ---- setup DDS ----
  si5351.init(SI5351_CRYSTAL_LOAD_8PF, 0, 0);
  si5351.set_correction(freq_cor, SI5351_PLL_INPUT_XO);
  si5351.drive_strength(SI5351_CLK0, SI5351_DRIVE_8MA);
  si5351_set_freq(0);

  display_dds_freq();
  
}

//------------
void loop() 
{
  float  f;
  int    i, rc;

  display_mode();
  
  while(true)
  {
    yield();
    check_mode();
  }
    
}

